//
//	LINKMM.CPP
//	Link Message to Message
//
//	Demonstrates how to create a message attachment which Microsoft
//	Exchange will recognize as a link to another message.
//
//	This is a command-line Windows program (!)  which takes a single argument:
//	the path to the message to be linked. It then creates a new message in the
//	Inbox conatining a link to that message. Use it like this:
//
char szUsage[] =
	"Usage: mlink message-path\r\n"
	"    Examples:\r\n"
	"    mlink Stationery\\Expense Report Template\r\n"
	"    mlink Public Folders\\Fishing Spots\\Big Two-hearted River\r\n"
	"Note: don't try to use quotes to preserve spaces.\r\n";
char szApp[] = "Link Message to Message";
//
//	For you BNF enthusiasts, it's [store] folder [folder...] message-subject
//	In other words, the store name is optional, but you must have at least one
//	folder.
//
//	The following information on links to messages and folders is not part
//	of the MAPI specification. It applies only to the Microsoft Exchange
//	client version 4.0. Other MAPI clients, as well as future versions
//	of the Microsoft Exchange client, may not be compatible with links
//	built according to the information supplied here.
//

//	Format string for test message body.
char szTextFmt[]	= "Link to message '%.256s'\r\n \r\n";
char szSubjectFmt[]	= "Link to '%.40s'";

#include <windows.h>
#include <mapix.h>
#include <mapiutil.h>
#include <string.h>
#include "_links.h"
#include "_resrc.h"

int CALLBACK
WinMain (HINSTANCE hInstance, HINSTANCE hPrevInst, LPSTR lpszCmd, int nCmdShow)
{
	HRESULT			hr;
	BOOL			fMAPIInit = FALSE;
	LPMAPISESSION	pses = NULL;
	LPSTR			szPath = NULL;
	LPSTR			szSubject = NULL;
	LPSTR			sz;
	LPMDB			pmdbDefault = NULL;
	LPMDB			pmdbLink = NULL;
	LPMAPIFOLDER	pfldInbox = NULL;
	LPMAPIFOLDER	pfldLink = NULL;
	LPMESSAGE		pmsg = NULL;
	LPENTRYLIST		pel = NULL;
	LPSPropValue	pval = NULL;
	LPSPropProblemArray pprobs = NULL;
	SPropValue		rgval[4];
	SYSTEMTIME		st;
	FILETIME		ft;
	int				ielParent;
	ULONG			ul;
	HICON			hicon = 0;
	ULONG			ichRender;
	char			szMsgText[512];
	char			szMsgSubject[64];
	
	//	Skip to the first argument.
	for (sz = lpszCmd; *sz && *sz == ' '; ++sz)
		;
	if (!*sz)
		goto usage;
	//	Copy the path because we're going to munge it.
	if (!(szPath = strdup(sz)))
		goto ret;
	//	Find the rightmost node of the path and treat it as the subject.
	for (sz = strpbrk(szPath, szPathSep); sz; sz = strpbrk(sz+1, szPathSep))
		szSubject = sz;
	if (!szSubject)
		goto usage;
	*szSubject++ = 0;

	//	Start up MAPI, log on, and open the Inbox.
	if (hr = OleInitialize(NULL))
		goto ret;
	if (hr = MAPIInitialize(NULL))
		goto ret;
	fMAPIInit = TRUE;
	if (hr = MAPILogonEx(0, NULL, NULL, MAPI_USE_DEFAULT | MAPI_NO_MAIL,
		&pses))
		goto ret;
	if (hr = OpenInbox(pses, &pmdbDefault, &pfldInbox))
		goto ret;

	//	Crack the folder path.
	if (hr = TraceFolderPath(pses, szPath, &pel, &pmdbLink))
		goto ret;

	//	Find the message in the indicated folder.
	ielParent = (int) pel->cValues - 1;
	if (hr = pmdbLink->OpenEntry(pel->lpbin[ielParent].cb,
		(LPENTRYID) pel->lpbin[ielParent].lpb,
		NULL, MAPI_DEFERRED_ERRORS, &ul, (LPUNKNOWN FAR *) &pfldLink))
		goto ret;
	if (hr = TraceMessage(pfldLink, szSubject, &pval))
	{
		goto ret;
	}

	//	Prepare certain cosmetic features
	hicon = LoadIcon(hInstance, MAKEINTRESOURCE(rsidMessageLink));
	wsprintf(szMsgText, szTextFmt, pval[ivalSubject_TM].Value.lpszA);
	{
		LPSTR	szT;

		szT = strrchr(szMsgText, ' ');
		ichRender = szT ? szT - szMsgText : 0xffffffff;
	}
	wsprintf(szMsgSubject, szSubjectFmt, pval[ivalSubject_TM].Value.lpszA);

	//	Create a new message in the Inbox, with recognizable subject
	//	and text.
	if (hr = pfldInbox->CreateMessage(NULL, MAPI_DEFERRED_ERRORS, &pmsg))
		goto ret;
	rgval[0].ulPropTag = PR_BODY;
	rgval[0].Value.lpszA = szMsgText;
	rgval[1].ulPropTag = PR_SUBJECT;
	rgval[1].Value.lpszA = szMsgSubject;
	rgval[2].ulPropTag = PR_MESSAGE_DELIVERY_TIME;
	GetSystemTime(&st);
	SystemTimeToFileTime(&st, &ft);
	rgval[2].Value.ft = ft;
	rgval[3].ulPropTag = PR_MESSAGE_CLASS;
	rgval[3].Value.lpszA = "IPM.Note";
	if (hr = pmsg->SetProps(4, rgval, &pprobs))
		goto ret;
	if (pprobs)
	{
		hr = ResultFromScode(pprobs->aProblem[0].scode);
		goto ret;
	}

	//	Lay down the message link.
	if (hr = WriteMessageLink(pmsg, &pel->lpbin[0], &pel->lpbin[ielParent],
		&pval[ivalEntryID_TM].Value.bin, pval[ivalSubject_TM].Value.lpszA,
			hicon, ichRender))
		goto ret;

	//	Save the message (don't send it).
	if (hr = pmsg->SaveChanges(0))
		goto ret;

ret:
	//	Clean up after ourselves.
	MAPIFreeBuffer(pel);
	MAPIFreeBuffer(pval);
	MAPIFreeBuffer(pprobs);
	UlRelease(pmsg);
	UlRelease(pfldLink);
	UlRelease(pmdbLink);
	UlRelease(pfldInbox);
	UlRelease(pmdbDefault);
	UlRelease(pses);
	if (fMAPIInit)
		MAPIUninitialize();
	OleUninitialize();
	if (szPath)
		free(szPath);
	if (hicon)
		DestroyIcon(hicon);
	
	//	Do the usual exit status thing: 
	//	0 => OK, nonzero => failure
	return ResultFromScode(hr) != S_OK;

usage:
	MessageBox(NULL, szUsage, szApp, MB_OK | MB_ICONSTOP);
	goto ret;
}
