//
//	LINKUTL.CPP
//
//	Common code for Microsoft Exchange message and folder linking samples.
//
//	The following information on links to messages and folders is not part
//	of the MAPI specification. It applies only to the Microsoft Exchange
//	client version 4.0. Other MAPI clients, as well as future versions
//	of the Microsoft Exchange client, may not be compatible with links
//	built according to the information supplied here.
//

#include <windows.h>
#include <mapix.h>
#include <mapiutil.h>
#include "_links.h"

//
//	Given a MAPI session in pses, OpenInbox returns the default message
//	store and inbox folder objects.
//

STDAPI	
OpenInbox(LPMAPISESSION pses, LPMDB FAR *ppmdb, LPMAPIFOLDER FAR *ppfld)
{
	HRESULT			hr;
	LPMDB			pmdb = NULL;
	LPMAPIFOLDER	pfld = NULL;
	LPMAPITABLE		ptbl = NULL;
	LPSRowSet		prows = NULL;
	LPSPropValue	pvalProp = NULL;
	ULONG			ul;
	ULONG			cbeid;
	LPENTRYID		peid = NULL;
	SPropValue		valDefStore;
	SRestriction	restDefStore;
	SPropertyRestriction restpropDefStore;
	static SizedSPropTagArray(2, ptaCol) =
				{ 2, { PR_DEFAULT_STORE, PR_ENTRYID} };

	//	Set up a restriction on the message stores table to identify the
	//	default message store.
	valDefStore.ulPropTag = PR_DEFAULT_STORE;
	valDefStore.dwAlignPad = 0;
	valDefStore.Value.b = TRUE;

	restpropDefStore.relop = RELOP_EQ;
	restpropDefStore.ulPropTag = PR_DEFAULT_STORE;
	restpropDefStore.lpProp = &valDefStore;

	restDefStore.rt = RES_PROPERTY;
	restDefStore.res.resProperty = restpropDefStore;

	//	Retrieve the message stores table.
	if (hr = pses->GetMsgStoresTable(0, &ptbl))
		goto ret;
	
	//	Find and open the default mesage store.
	hr = HrQueryAllRows(ptbl, (LPSPropTagArray) &ptaCol, &restDefStore, 
		NULL, 0, &prows);
	if (hr)
		goto ret;
	if (prows == NULL || prows->cRows == 0
		|| prows->aRow[0].lpProps[1].ulPropTag != PR_ENTRYID)
	{
		hr = ResultFromScode(MAPI_E_NOT_FOUND);
		goto ret;
	}
	hr = pses->OpenMsgStore(0,
		prows->aRow[0].lpProps[1].Value.bin.cb,
		(LPENTRYID) prows->aRow[0].lpProps[1].Value.bin.lpb,
		NULL, MDB_WRITE | MDB_NO_DIALOG | MAPI_DEFERRED_ERRORS, &pmdb);
	if (HR_FAILED(hr))
		goto ret;

	//	Locate and open the Inbox folder.
	if (hr = pmdb->GetReceiveFolder("IPM", 0, &cbeid, &peid, NULL))
		goto ret;
	if (hr = pmdb->OpenEntry(cbeid, peid, NULL, 
		MAPI_DEFERRED_ERRORS | MAPI_MODIFY, &ul, (LPUNKNOWN FAR *)&pfld))
		goto ret;

	//	Hand back our output parameters, and prevent their release.
	*ppmdb = pmdb;
	*ppfld = pfld;
	pmdb = NULL;
	pfld = NULL;
					
ret:
	MAPIFreeBuffer(peid);
	FreeProws(prows);
	UlRelease(ptbl);

	return hr;
}

//
//	Given an icon handle, GetIconMetafile returns a byte stream
//	in *ppb with the Windows metafile corresponding to the icon.
//	The byte count is returned in *pcb.
//
//	The memory that holds the metafile stream is allocated using
//	MAPIAllocateBuffer and must be freed using MAPIFreeBuffer.
//

STDAPI
GetIconMetafile(HICON hicon, ULONG FAR *pcb, LPBYTE FAR *ppb)
{
	HRESULT		hr;
	HDC			hdc;
	HANDLE		hmf;
	ULONG		cb;
	ULONG		cbT;
	LPBYTE		pb;

	//	Render the icon into a metafile.
	hdc = CreateMetaFile(NULL);
	if (!hdc)
		goto fail;
	DrawIcon(hdc, 0, 0, hicon);
	hmf = CloseMetaFile (hdc);

	//	We are done with the device context now; check for success.
	hdc = NULL;
	if (!hmf)
		goto fail;

	//	Dump the metafile bits into a buffer.
#ifdef _WIN32
	cbT = GetMetaFileBitsEx(hmf, 0, NULL);

	if (hr = ResultFromScode(MAPIAllocateBuffer(cbT, (LPVOID *) &pb)))
		goto ret;

	cb = GetMetaFileBitsEx(hmf, cbT, pb);
	if (cb != cbT)
		goto fail;
#else // WIN16
	{
		//	Same logic as Win32, different API calls
		HANDLE 	hmfT = GetMetaFileBits(hmf);
		LPBYTE	pbmf;

		if (!hmfT)
			goto fail;
		hmf = NULL;

		pbmf = GlobalLock(hmfT);
		cb = GlobalSize(hmfT);

		if (hr = ResultFromScode(MAPIAllocateBuffer(cb, (LPVOID *) &pb)))
		{
			GlobalUnlock (hmfT);
			GlobalFree (hmfT);
			goto ret;
		}

		memcpy(pb, pbmf, (size_t) cb);

		GlobalUnlock (hmfT);
		GlobalFree (hmfT);
	}
#endif	// _WIN32

	//	Everything worked, now return our results.
	*pcb = cb;
	*ppb = pb;

ret:
	return hr;

fail:
	hr = ResultFromScode(MAPI_E_CALL_FAILED);
	goto ret;
}

//
//	GetIconStringMetafile creates a Windows metafile from an icon and a
//	string. It uses the string as a caption for the icon. The metafile
//	is returned as a byte stream in *ppb, with the byte count in *pcb.
//
//	The memory that holds the metafile stream is allocated using
//	MAPIAllocateBuffer and must be freed using MAPIFreeBuffer.
//

STDAPI
GetIconStringMetafile(HICON hicon, LPSTR sz, ULONG FAR *pcb, LPBYTE FAR *ppb)
{
	HRESULT		hr;
	HGLOBAL		hglb = 0;
	LPBYTE		pbmf = NULL;
	LPMETAFILEPICT pmfp = NULL;
	ULONG		cb;
	LPBYTE		pb = NULL;
#define cchTitle 256
	OLECHAR		szTitle[cchTitle];
	int			cch;

	//	Convert and limit length of the title. If the conversion fails,
	//	the title will be blank.
	cch = lstrlenA(sz);
	if (cch > cchTitle - 1)
		cch = cchTitle - 1;
#ifdef	_WIN32
	MultiByteToWideChar(CP_ACP, 0, sz, cch, szTitle, cchTitle);
#else
	lstrcpyn(szTitle, sz, cchTitle - 1);
	szTitle[cchTitle - 1] = 0;
#endif

	//	Note: despite its name, this handy function does not appear to
	//	require OleInitialize().
	if (!(hglb = OleMetafilePictFromIconAndLabel(hicon, szTitle,
		NULL, 0)))
		goto fail;
	if (!(pmfp = (LPMETAFILEPICT) GlobalLock(hglb)))
		goto fail;

	pbmf = (LPBYTE) GlobalLock(pmfp->hMF);
	cb = GlobalSize(pmfp->hMF);
	if (hr = ResultFromScode(MAPIAllocateBuffer(cb, (LPVOID FAR *) &pb)))
		goto ret;
	memcpy(pb, pbmf, (size_t) cb);

	*pcb = cb;
	*ppb = pb;

ret:
	if (pmfp)
	{
		if (pbmf)
			GlobalUnlock(pmfp->hMF);
		GlobalFree(pmfp->hMF);
		GlobalUnlock(hglb);
	}
	if (hglb)
		GlobalFree(hglb);
	return hr;

fail:
	hr = ResultFromScode(MAPI_E_CALL_FAILED);
	goto ret;
}

//
//	SetAttachmentIcon simply takes an icon, creates the corresponding
//	metafile, and makes it the rendering for an attached file.
//

STDAPI
SetAttachmentIcon(HICON hicon, LPATTACH patch)
{
	HRESULT		hr;
	SPropValue	val;

	memset(&val, 0, sizeof(SPropValue));

	//	Convert the icon to a metafile.
	if (hr = GetIconMetafile(hicon, &val.Value.bin.cb, &val.Value.bin.lpb))
		goto ret;

	//	Create the attachment property.
	val.ulPropTag = PR_ATTACH_RENDERING;
	hr = HrSetOneProp((LPMAPIPROP) patch, &val);

ret:
	MAPIFreeBuffer(val.Value.bin.lpb);

	return hr;
}

//
//	SetStgRendering takes an icon and an optional caption (string),
//	turns them into a metafile, and writes the metafile out to the
//	stream named szStreamName in the stuctured storage pstg.
//

STDAPI
SetStgRendering(HICON hicon, LPSTR sz, LPSTORAGE pstg, LPOLESTR szStreamName)
{
	HRESULT		hr;
	ULONG		cb;
	LPBYTE		pb = NULL;

	if (sz)
	{
		//	If we have a caption, include it in the rendering.
		if (hr = GetIconStringMetafile(hicon, sz, &cb, &pb))
			goto ret;
	}
	else
	{
		//	No caption, just render the icon.
		if (hr = GetIconMetafile(hicon, &cb, &pb))
			goto ret;
	}

	//	Write the resulting metafile onto the storage.
	if (hr = PutStgStream(pstg, szStreamName, cb, pb))
		goto ret;

ret:
	MAPIFreeBuffer(pb);

	return hr;
}

//
//	PutStgStream creates the stream szStreamName in the structured
//	storage pstg, then writes data there. The address of the data is
//	pb, and the byte count is cb.
//
//	The storage is not committed.
//

STDAPI
PutStgStream(LPSTORAGE pstg, LPOLESTR szStreamName, ULONG cb, LPBYTE pb)
{
	HRESULT		hr;
	LPSTREAM	pstm = NULL;

	// Create the stream and write the data to it.
	if (!HR_FAILED(hr = pstg->CreateStream(szStreamName,
		STGM_READWRITE | STGM_CREATE | STGM_SHARE_EXCLUSIVE, 0, 0, &pstm)))
	{
		hr = pstm->Write(pb, cb, NULL);
		//	fall through
	}

	UlRelease(pstm);
	return hr;
}

